@@ -1,14 +1,20 @@
 # -*- coding: utf-8 -*-
 
+from copy import deepcopy
+
+from hypothesis import assume
+from hypothesis.stateful import Bundle, RuleBasedStateMachine, rule
+import hypothesis.strategies as st
+
 import pytest
 
 import vdirsyncer.exceptions as exceptions
 from vdirsyncer.storage.base import Item
-from vdirsyncer.storage.memory import MemoryStorage
+from vdirsyncer.storage.memory import MemoryStorage, _random_string
 from vdirsyncer.sync import BothReadOnly, IdentConflict, StorageEmpty, \
     SyncConflict, sync
 
-from . import assert_item_equals, blow_up, normalize_item
+from . import assert_item_equals, blow_up, uid_strategy
 
 
 def empty_storage(x):
@@ -41,6 +47,7 @@ def test_missing_status():
 def test_missing_status_and_different_items():
     a = MemoryStorage()
     b = MemoryStorage()
+
     status = {}
     item1 = Item(u'UID:1\nhaha')
     item2 = Item(u'UID:1\nhoho')
@@ -54,6 +61,23 @@ def test_missing_status_and_different_items():
     assert_item_equals(item1, a.get('1')[0])
 
 
+def test_read_only_and_prefetch():
+    a = MemoryStorage()
+    b = MemoryStorage()
+    b.read_only = True
+
+    status = {}
+    item1 = Item(u'UID:1\nhaha')
+    item2 = Item(u'UID:2\nhoho')
+    a.upload(item1)
+    a.upload(item2)
+
+    sync(a, b, status, force_delete=True)
+    sync(a, b, status, force_delete=True)
+
+    assert list(a.list()) == list(b.list()) == []
+
+
 def test_upload_and_update():
     a = MemoryStorage(fileext='.a')
     b = MemoryStorage(fileext='.b')
@@ -108,7 +132,13 @@ def test_already_synced():
     a.upload(item)
     b.upload(item)
     status = {
-        '1': ('1.a', a.get('1.a')[1], '1.b', b.get('1.b')[1])
+        '1': ({
+            'href': '1.a',
+            'etag': a.get('1.a')[1]
+        }, {
+            'href': '1.b',
+            'etag': b.get('1.b')[1]
+        })
     }
     old_status = dict(status)
     a.update = b.update = a.upload = b.upload = \
@@ -138,7 +168,7 @@ def test_conflict_resolution_both_etags_new(winning_storage):
     item_a, _ = a.get(href_a)
     item_b, _ = b.get(href_b)
     assert_item_equals(item_a, item_b)
-    n = normalize_item(item_a)
+    n = item_a.raw.splitlines()
     assert u'UID:1' in n
     assert u'item {}'.format(winning_storage) in n
 
@@ -176,8 +206,15 @@ def test_conflict_resolution_new_etags_without_changes():
     href_a, etag_a = a.upload(item)
     href_b, etag_b = b.upload(item)
     status = {'1': (href_a, 'BOGUS_a', href_b, 'BOGUS_b')}
+
     sync(a, b, status)
-    assert status == {'1': (href_a, etag_a, href_b, etag_b)}
+
+    (ident, (status_a, status_b)), = status.items()
+    assert ident == '1'
+    assert status_a['href'] == href_a
+    assert status_a['etag'] == etag_a
+    assert status_b['href'] == href_b
+    assert status_b['etag'] == etag_b
 
 
 def test_uses_get_multi(monkeypatch):
@@ -189,6 +226,7 @@ def test_uses_get_multi(monkeypatch):
     old_get = MemoryStorage.get
 
     def get_multi(self, hrefs):
+        hrefs = list(hrefs)
         get_multi_calls.append(hrefs)
         for href in hrefs:
             item, etag = old_get(self, href)
@@ -234,6 +272,18 @@ def test_no_uids():
     assert a_items == b_items == {u'ASDF', u'FOOBAR'}
 
 
+def test_changed_uids():
+    a = MemoryStorage()
+    b = MemoryStorage()
+    href_a, etag_a = a.upload(Item(u'UID:A-ONE'))
+    href_b, etag_b = b.upload(Item(u'UID:B-ONE'))
+    status = {}
+    sync(a, b, status)
+
+    a.update(href_a, Item(u'UID:A-TWO'), etag_a)
+    sync(a, b, status)
+
+
 def test_both_readonly():
     a = MemoryStorage(read_only=True)
     b = MemoryStorage(read_only=True)
@@ -292,9 +342,134 @@ def test_moved_href():
     sync(a, b, status)
 
     b.items['lol'] = b.items.pop('haha')
-    a.delete = a.update = a.upload = blow_up
+
+    # The sync algorithm should prefetch `lol`, see that it's the same ident
+    # and not do anything else.
+    a.get_multi = blow_up  # Absolutely no prefetch on A
+    # No actual sync actions
+    a.delete = a.update = a.upload = b.delete = b.update = b.upload = blow_up
 
     sync(a, b, status)
     assert len(status) == 1
     assert len(list(a.list())) == len(list(b.list())) == 1
-    assert status['haha'][2] == 'haha'
+    assert status['haha'][1]['href'] == 'lol'
+    old_status = deepcopy(status)
+
+    # Further sync should be a noop. Not even prefetching should occur.
+    b.get_multi = blow_up
+
+    sync(a, b, status)
+    assert old_status == status
+    assert len(list(a.list())) == len(list(b.list())) == 1
+
+
+def test_bogus_etag_change():
+    '''Assert that sync algorithm is resilient against etag changes if content
+    didn\'t change.
+
+    In this particular case we test a scenario where both etags have been
+    updated, but only one side actually changed its item content.
+    '''
+    a = MemoryStorage()
+    b = MemoryStorage()
+    status = {}
+    href_a, etag_a = a.upload(Item(u'UID:ASDASD'))
+    sync(a, b, status)
+    assert len(status) == len(list(a.list())) == len(list(b.list())) == 1
+
+    (href_b, etag_b), = b.list()
+    a.update(href_a, Item(u'UID:ASDASD'), etag_a)
+    b.update(href_b, Item(u'UID:ASDASD\nACTUALCHANGE:YES'), etag_b)
+
+    b.delete = b.update = b.upload = blow_up
+
+    sync(a, b, status)
+    assert len(status) == len(list(a.list())) == len(list(b.list())) == 1
+    assert a.get(href_a)[0].raw == u'UID:ASDASD\nACTUALCHANGE:YES'
+
+
+def test_unicode_hrefs():
+    a = MemoryStorage()
+    b = MemoryStorage()
+    status = {}
+    href, etag = a.upload(Item(u'UID:äää'))
+    sync(a, b, status)
+
+
+class SyncMachine(RuleBasedStateMachine):
+    Status = Bundle('status')
+    Storage = Bundle('storage')
+
+    @staticmethod
+    def _get_items(storage):
+        return sorted(item.raw for etag, item in storage.items.values())
+
+    @rule(target=Storage, read_only=st.booleans(), flaky_etags=st.booleans())
+    def newstorage(self, read_only, flaky_etags):
+        s = MemoryStorage()
+        s.read_only = read_only
+        if flaky_etags:
+            def get(href):
+                _, item = s.items[href]
+                etag = _random_string()
+                s.items[href] = etag, item
+                return item, etag
+            s.get = get
+
+        return s
+
+    @rule(target=Status)
+    def newstatus(self):
+        return {}
+
+    @rule(target=Storage, storage=Storage,
+          uid=uid_strategy,
+          etag=st.text())
+    def upload(self, storage, uid, etag):
+        item = Item(u'UID:{}'.format(uid))
+        storage.items[uid] = (etag, item)
+        return storage
+
+    @rule(target=Storage, storage=Storage, href=st.text())
+    def delete(self, storage, href):
+        storage.items.pop(href, None)
+        return storage
+
+    @rule(
+        target=Status, status=Status,
+        a=Storage, b=Storage,
+        force_delete=st.booleans(),
+        conflict_resolution=st.one_of((st.just('a wins'), st.just('b wins')))
+    )
+    def sync(self, status, a, b, force_delete, conflict_resolution):
+        old_items_a = self._get_items(a)
+        old_items_b = self._get_items(b)
+
+        try:
+            # If one storage is read-only, double-sync because changes don't
+            # get reverted immediately.
+            for _ in range(2 if a.read_only or b.read_only else 1):
+                sync(a, b, status,
+                     force_delete=force_delete,
+                     conflict_resolution=conflict_resolution)
+        except BothReadOnly:
+            assert a.read_only and b.read_only
+            assume(False)
+        except StorageEmpty:
+            if force_delete:
+                raise
+            else:
+                assert not list(a.list()) or not list(b.list())
+                return status
+
+        items_a = self._get_items(a)
+        items_b = self._get_items(b)
+
+        assert items_a == items_b
+        assert items_a == old_items_a or not a.read_only
+        assert items_b == old_items_b or not b.read_only
+
+        return status
+
+
+TestSyncMachine = SyncMachine.TestCase
