@@ -1,59 +1,80 @@
 # -*- coding: utf-8 -*-
 
 import datetime
+import logging
 
-from lxml import etree
+import xml.etree.ElementTree as etree
 
 import requests
 from requests.exceptions import HTTPError
 
-from .base import Item, Storage
+from .base import Item, Storage, normalize_meta_value
 from .http import HTTP_STORAGE_PARAMETERS, USERAGENT, prepare_auth, \
     prepare_client_cert, prepare_verify
-from .. import exceptions, log, utils
-from ..utils.compat import text_type
+from .. import exceptions, utils
+from ..utils.compat import PY2, getargspec_ish, text_type, to_native
 
 
-dav_logger = log.get(__name__)
+dav_logger = logging.getLogger(__name__)
 
 CALDAV_DT_FORMAT = '%Y%m%dT%H%M%SZ'
 
+_path_reserved_chars = frozenset(utils.compat.urlquote(x, '')
+                                 for x in "/?#[]!$&'()*+,;=")
+
+
+def _contains_quoted_reserved_chars(x):
+    for y in _path_reserved_chars:
+        if y in x:
+            dav_logger.debug('Unsafe character: {!r}'.format(y))
+            return True
+    return False
+
 
 def _normalize_href(base, href):
     '''Normalize the href to be a path only relative to hostname and
     schema.'''
+    orig_href = href
+    base = to_native(base, 'utf-8')
+    href = to_native(href, 'utf-8')
     if not href:
         raise ValueError(href)
+
     x = utils.compat.urlparse.urljoin(base, href)
     x = utils.compat.urlparse.urlsplit(x).path
+
+    # Encoding issues:
+    # - https://github.com/owncloud/contacts/issues/581
+    # - https://github.com/Kozea/Radicale/issues/298
+    old_x = None
+    while old_x is None or x != old_x:
+        if _contains_quoted_reserved_chars(x):
+            break
+        old_x = x
+        x = utils.compat.urlunquote(x)
+
+    x = utils.compat.urlquote(x, '/@%:')
+
+    if orig_href == x:
+        dav_logger.debug('Already normalized: {!r}'.format(x))
+    else:
+        dav_logger.debug('Normalized URL from {!r} to {!r}'
+                         .format(orig_href, x))
+
     return x
 
 
-def _encode_href(x):
-    return utils.compat.urlquote(x, '/@')
-
-
-def _decode_href(x):
-    return utils.compat.urlunquote(x)
-
-
 class InvalidXMLResponse(exceptions.InvalidResponse):
     pass
 
 
 def _parse_xml(content):
-    p = etree.XMLParser(recover=True)
-    rv = etree.XML(content, parser=p)
-    if rv is None:
+    try:
+        return etree.XML(content)
+    except etree.ParseError as e:
         raise InvalidXMLResponse('Invalid XML encountered: {}\n'
                                  'Double-check the URLs in your config.'
-                                 .format(p.error_log))
-    if p.error_log:
-        dav_logger.warning('Partially invalid XML response, some of your '
-                           'items may be corrupted. Check the debug log and '
-                           'consider switching servers. ({})'
-                           .format(p.error_log))
-    return rv
+                                 .format(e))
 
 
 def _merge_xml(items):
@@ -76,12 +97,8 @@ def _fuzzy_matches_mimetype(strict, weak):
     return False
 
 
-def _get_collection_from_url(url):
-    _, collection = url.rstrip('/').rsplit('/', 1)
-    return collection
-
-
 class Discover(object):
+    _namespace = None
     _resourcetype = None
     _homeset_xml = None
     _homeset_tag = None
@@ -90,13 +107,21 @@ class Discover(object):
     <d:propfind xmlns:d="DAV:">
         <d:prop>
             <d:resourcetype />
-            <d:displayname />
         </d:prop>
     </d:propfind>
     """
 
-    def __init__(self, session):
+    def __init__(self, session, kwargs):
+        if kwargs.pop('collection', None) is not None:
+            raise TypeError('collection argument must not be given.')
+
         self.session = session
+        self.kwargs = kwargs
+
+    @staticmethod
+    def _get_collection_from_url(url):
+        _, collection = url.rstrip('/').rsplit('/', 1)
+        return utils.compat.urlunquote(collection)
 
     def find_dav(self):
         try:
@@ -113,7 +138,11 @@ class Discover(object):
 
     def find_principal(self, url=None):
         if url is None:
-            url = self.find_dav()
+            try:
+                return self.find_principal('')
+            except (HTTPError, exceptions.Error):
+                return self.find_principal(self.find_dav())
+
         headers = self.session.get_default_headers()
         headers['Depth'] = 0
         body = """
@@ -142,9 +171,9 @@ class Discover(object):
 
         root = etree.fromstring(response.content)
         # Better don't do string formatting here, because of XML namespaces
-        rv = root.find('.//' + self._homeset_tag + '/{*}href')
+        rv = root.find('.//' + self._homeset_tag + '/{DAV:}href')
         if rv is None:
-            raise InvalidXMLResponse()
+            raise InvalidXMLResponse('Couldn\'t find home-set.')
         return utils.compat.urlparse.urljoin(response.url, rv.text)
 
     def find_collections(self, url=None):
@@ -157,35 +186,79 @@ class Discover(object):
         root = _parse_xml(r.content)
         done = set()
         for response in root.findall('{DAV:}response'):
-            props = _merge_xml(response.findall('{*}propstat/{*}prop'))
-            if props.find('{*}resourcetype/{*}' + self._resourcetype) is None:
+            props = _merge_xml(response.findall('{DAV:}propstat/{DAV:}prop'))
+            if props.find('{DAV:}resourcetype/' + self._resourcetype) is None:
                 continue
 
-            displayname = getattr(props.find('{*}displayname'), 'text', '')
-            href = response.find('{*}href')
+            href = response.find('{DAV:}href')
             if href is None:
-                raise InvalidXMLResponse()
+                raise InvalidXMLResponse('Missing href tag for collection '
+                                         'props.')
             href = utils.compat.urlparse.urljoin(r.url, href.text)
             if href not in done:
                 done.add(href)
-                yield {'href': href, 'displayname': displayname}
+                yield {'href': href}
 
     def discover(self):
-        for x in self.find_collections():
-            yield x
+        for c in self.find_collections():
+            url = c['href']
+            collection = self._get_collection_from_url(url)
+            storage_args = dict(self.kwargs)
+            storage_args.update({'url': url, 'collection': collection})
+            yield storage_args
+
+    def create(self, collection):
+        if collection is None:
+            collection = self._get_collection_from_url(self.kwargs['url'])
+
+        for c in self.discover():
+            if c['collection'] == collection:
+                return c
+
+        home = self.find_home()
+        url = utils.compat.urlparse.urljoin(
+            home,
+            utils.compat.urlquote(collection, '/@')
+        )
 
-        # Another one of Radicale's specialties: Discovery is broken (returning
-        # completely wrong URLs at every stage) as of version 0.9.
-        # https://github.com/Kozea/Radicale/issues/196
         try:
-            for x in self.find_collections(''):
-                yield x
-        except (InvalidXMLResponse, HTTPError, exceptions.Error):
-            pass
+            url = self._create_collection_impl(url)
+        except HTTPError as e:
+            raise NotImplementedError(e)
+        else:
+            rv = dict(self.kwargs)
+            rv['collection'] = collection
+            rv['url'] = url
+            return rv
+
+    def _create_collection_impl(self, url):
+        data = '''<?xml version="1.0" encoding="utf-8" ?>
+            <D:mkcol xmlns:D="DAV:">
+                <D:set>
+                    <D:prop>
+                        <D:resourcetype>
+                            <D:collection/>
+                            {}
+                        </D:resourcetype>
+                    </D:prop>
+                </D:set>
+            </D:mkcol>
+        '''.format(
+            to_native(etree.tostring(etree.Element(self._resourcetype)))
+        )
+
+        response = self.session.request(
+            'MKCOL',
+            url,
+            data=data,
+            headers=self.session.get_default_headers(),
+        )
+        return response.url
 
 
 class CalDiscover(Discover):
-    _resourcetype = 'calendar'
+    _namespace = 'urn:ietf:params:xml:ns:caldav'
+    _resourcetype = '{%s}calendar' % _namespace
     _homeset_xml = """
     <d:propfind xmlns:d="DAV:" xmlns:c="urn:ietf:params:xml:ns:caldav">
         <d:prop>
@@ -193,12 +266,13 @@ class CalDiscover(Discover):
         </d:prop>
     </d:propfind>
     """
-    _homeset_tag = '{*}calendar-home-set'
+    _homeset_tag = '{%s}calendar-home-set' % _namespace
     _well_known_uri = '/.well-known/caldav/'
 
 
 class CardDiscover(Discover):
-    _resourcetype = 'addressbook'
+    _namespace = 'urn:ietf:params:xml:ns:carddav'
+    _resourcetype = '{%s}addressbook' % _namespace
     _homeset_xml = """
     <d:propfind xmlns:d="DAV:" xmlns:c="urn:ietf:params:xml:ns:carddav">
         <d:prop>
@@ -206,7 +280,7 @@ class CardDiscover(Discover):
         </d:prop>
     </d:propfind>
     """
-    _homeset_tag = '{*}addressbook-home-set'
+    _homeset_tag = '{%s}addressbook-home-set' % _namespace
     _well_known_uri = '/.well-known/carddav/'
 
 
@@ -215,30 +289,36 @@ class DavSession(object):
     A helper class to connect to DAV servers.
     '''
 
+    @classmethod
+    def init_and_remaining_args(cls, **kwargs):
+        argspec = getargspec_ish(cls.__init__)
+        self_args, remainder = \
+            utils.split_dict(kwargs, argspec.args.__contains__)
+
+        return cls(**self_args), remainder
+
     def __init__(self, url, username='', password='', verify=True, auth=None,
                  useragent=USERAGENT, verify_fingerprint=None,
                  auth_cert=None):
-        if username and not password:
-            from ..utils.password import get_password
-            password = get_password(username, url)
-
         self._settings = {
-            'auth': prepare_auth(auth, username, password),
             'cert': prepare_client_cert(auth_cert),
+            'auth': prepare_auth(auth, username, password)
         }
         self._settings.update(prepare_verify(verify, verify_fingerprint))
 
         self.useragent = useragent
         self.url = url.rstrip('/') + '/'
-        self.parsed_url = utils.compat.urlparse.urlparse(self.url)
-        self._session = None
+
+        self._session = requests.session()
+
+    @utils.cached_property
+    def parsed_url(self):
+        return utils.compat.urlparse.urlparse(self.url)
 
     def request(self, method, path, **kwargs):
         url = self.url
         if path:
             url = utils.compat.urlparse.urljoin(self.url, path)
-        if self._session is None:
-            self._session = requests.session()
 
         more = dict(self._settings)
         more.update(kwargs)
@@ -256,8 +336,6 @@ class DavStorage(Storage):
     __doc__ = '''
     :param url: Base URL or an URL to a collection.
     ''' + HTTP_STORAGE_PARAMETERS + '''
-    :param unsafe_href_chars: Replace the given characters when generating
-        hrefs. Defaults to ``'@'``.
 
     .. note::
 
@@ -275,101 +353,45 @@ class DavStorage(Storage):
     get_multi_data_query = None
     # The Discover subclass to use
     discovery_class = None
+    # The DavSession class to use
+    session_class = DavSession
 
-    _session = None
     _repr_attributes = ('username', 'url')
 
-    def __init__(self, url, username='', password='', verify=True, auth=None,
-                 useragent=USERAGENT, unsafe_href_chars='@',
-                 verify_fingerprint=None, auth_cert=None, **kwargs):
+    _property_table = {
+        'displayname': ('displayname', 'DAV:'),
+    }
+
+    def __init__(self, **kwargs):
+        # defined for _repr_attributes
+        self.username = kwargs.get('username')
+        self.url = kwargs.get('url')
+
+        self.session, kwargs = \
+            self.session_class.init_and_remaining_args(**kwargs)
         super(DavStorage, self).__init__(**kwargs)
 
-        url = url.rstrip('/') + '/'
-        self.session = DavSession(url, username, password, verify, auth,
-                                  useragent, verify_fingerprint,
-                                  auth_cert)
-        self.unsafe_href_chars = unsafe_href_chars
-
-        # defined for _repr_attributes
-        self.username = username
-        self.url = url
-
-    @classmethod
-    def _get_session(cls, **kwargs):
-        discover_args, _ = utils.split_dict(kwargs, lambda key: key in (
-            'url', 'username', 'password', 'verify', 'auth', 'useragent',
-            'verify_fingerprint', 'auth_cert',
-        ))
-        return DavSession(**discover_args)
+    if not PY2:
+        import inspect
+        __init__.__signature__ = inspect.signature(session_class.__init__)
 
     @classmethod
     def discover(cls, **kwargs):
-        if kwargs.pop('collection', None) is not None:
-            raise TypeError('collection argument must not be given.')
-
-        d = cls.discovery_class(cls._get_session(**kwargs))
-        for c in d.discover():
-            url = c['href']
-            collection = _get_collection_from_url(url)
-            storage_args = dict(kwargs)
-            storage_args.update({'url': url, 'collection': collection,
-                                 'collection_human': c['displayname']})
-            yield storage_args
+        session, _ = cls.session_class.init_and_remaining_args(**kwargs)
+        d = cls.discovery_class(session, kwargs)
+        return d.discover()
 
     @classmethod
     def create_collection(cls, collection, **kwargs):
-        if collection is None:
-            collection = _get_collection_from_url(kwargs['url'])
-        session = cls._get_session(**kwargs)
-        d = cls.discovery_class(session)
-
-        for c in cls.discover(**kwargs):
-            if c['collection'] == collection:
-                return c
-
-        home = d.find_home()
-        url = '{}/{}'.format(home.rstrip('/'), collection)
-
-        try:
-            url = cls._create_collection_impl(d, url, kwargs)
-        except HTTPError as e:
-            raise NotImplementedError(e)
-        else:
-            rv = dict(kwargs)
-            rv['collection'] = collection
-            rv['url'] = url
-            return rv
-
-    @classmethod
-    def _create_collection_impl(cls, d, url, kwargs):
-        data = '''<?xml version="1.0" encoding="utf-8" ?>
-            <D:mkcol xmlns:D="DAV:" xmlns:C="{}">
-                <D:set>
-                    <D:prop>
-                        <D:resourcetype>
-                            <D:collection/>
-                            <C:{}/>
-                        </D:resourcetype>
-                    </D:prop>
-                </D:set>
-            </D:mkcol>
-        '''.format(cls._dav_namespace, cls._dav_resourcetype)
-
-        response = d.session.request(
-            'MKCOL',
-            url,
-            data=data,
-            headers=d.session.get_default_headers(),
-        )
-        return response.url
+        session, _ = cls.session_class.init_and_remaining_args(**kwargs)
+        d = cls.discovery_class(session, kwargs)
+        return d.create(collection)
 
     def _normalize_href(self, *args, **kwargs):
         return _normalize_href(self.session.url, *args, **kwargs)
 
     def _get_href(self, item):
-        href = item.ident
-        for char in self.unsafe_href_chars:
-            href = href.replace(char, '_')
+        href = utils.generate_href(item.ident)
         return self._normalize_href(href + self.fileext)
 
     def _is_item_mimetype(self, mimetype):
@@ -386,7 +408,7 @@ class DavStorage(Storage):
         for href in hrefs:
             if href != self._normalize_href(href):
                 raise exceptions.NotFoundError(href)
-            href_xml.append('<D:href>{}</D:href>'.format(_encode_href(href)))
+            href_xml.append('<D:href>{}</D:href>'.format(href))
         if not href_xml:
             return ()
 
@@ -439,15 +461,20 @@ class DavStorage(Storage):
 
         response = self.session.request(
             'PUT',
-            _encode_href(href),
+            href,
             data=item.raw.encode('utf-8'),
             headers=headers
         )
         etag = response.headers.get('etag', None)
-        href = self._normalize_href(_decode_href(response.url))
+        href = self._normalize_href(response.url)
         if not etag:
+            # The server violated the RFC and didn't send an etag. This is
+            # technically a race-condition, but too many popular servers do it.
+            #
+            # ownCloud: https://github.com/owncloud/contacts/issues/920
+            dav_logger.debug('Server did not send etag, fetching {!r}'
+                             .format(href))
             item2, etag = self.get(href)
-            assert item2.uid == item.uid
         return href, etag
 
     def update(self, href, item, etag):
@@ -469,11 +496,11 @@ class DavStorage(Storage):
 
         self.session.request(
             'DELETE',
-            _encode_href(href),
+            href,
             headers=headers
         )
 
-    def _parse_prop_responses(self, root, decoding_rounds=1):
+    def _parse_prop_responses(self, root):
         hrefs = set()
         for response in root.iter('{DAV:}response'):
             href = response.find('{DAV:}href')
@@ -482,15 +509,13 @@ class DavStorage(Storage):
                 continue
 
             href = self._normalize_href(href.text)
-            for i in range(decoding_rounds):
-                href = _decode_href(href)
 
             if href in hrefs:
                 # Servers that send duplicate hrefs:
                 # - Zimbra
-                #   https://github.com/untitaker/vdirsyncer/issues/88
+                #   https://github.com/pimutils/vdirsyncer/issues/88
                 # - Davmail
-                #   https://github.com/untitaker/vdirsyncer/issues/144
+                #   https://github.com/pimutils/vdirsyncer/issues/144
                 dav_logger.warning('Skipping identical href: {!r}'
                                    .format(href))
                 continue
@@ -538,17 +563,79 @@ class DavStorage(Storage):
             '''
 
         # We use a PROPFIND request instead of addressbook-query due to issues
-        # with Zimbra. See https://github.com/untitaker/vdirsyncer/issues/83
+        # with Zimbra. See https://github.com/pimutils/vdirsyncer/issues/83
         response = self.session.request('PROPFIND', '', data=data,
                                         headers=headers)
         root = _parse_xml(response.content)
 
-        # Decode twice because ownCloud encodes twice.
-        # See https://github.com/owncloud/contacts/issues/581
-        rv = self._parse_prop_responses(root, decoding_rounds=2)
+        rv = self._parse_prop_responses(root)
         for href, etag, prop in rv:
             yield href, etag
 
+    def get_meta(self, key):
+        try:
+            tagname, namespace = self._property_table[key]
+        except KeyError:
+            raise exceptions.UnsupportedMetadataError()
+
+        xpath = '{%s}%s' % (namespace, tagname)
+        data = '''<?xml version="1.0" encoding="utf-8" ?>
+            <D:propfind xmlns:D="DAV:">
+                <D:prop>
+                    {}
+                </D:prop>
+            </D:propfind>
+        '''.format(
+            to_native(etree.tostring(etree.Element(xpath)))
+        )
+
+        headers = self.session.get_default_headers()
+        headers['Depth'] = 0
+
+        response = self.session.request(
+            'PROPFIND', '',
+            data=data, headers=headers
+        )
+
+        root = _parse_xml(response.content)
+
+        for prop in root.findall('.//' + xpath):
+            text = normalize_meta_value(getattr(prop, 'text', None))
+            if text:
+                return text
+        return u''
+
+    def set_meta(self, key, value):
+        try:
+            tagname, namespace = self._property_table[key]
+        except KeyError:
+            raise exceptions.UnsupportedMetadataError()
+
+        lxml_selector = '{%s}%s' % (namespace, tagname)
+        element = etree.Element(lxml_selector)
+        element.text = normalize_meta_value(value)
+
+        data = '''<?xml version="1.0" encoding="utf-8" ?>
+            <D:propertyupdate xmlns:D="DAV:">
+                <D:set>
+                    <D:prop>
+                        {}
+                    </D:prop>
+                </D:set>
+            </D:propertyupdate>
+        '''.format(to_native(etree.tostring(element)))
+
+        self.session.request(
+            'PROPPATCH', '',
+            data=data, headers=self.session.get_default_headers()
+        )
+
+        # XXX: Response content is currently ignored. Though exceptions are
+        # raised for HTTP errors, a multistatus with errorcodes inside is not
+        # parsed yet. Not sure how common those are, or how they look like. It
+        # might be easier (and safer in case of a stupid server) to just issue
+        # a PROPFIND to see if the value got actually set.
+
 
 class CaldavStorage(DavStorage):
 
@@ -585,8 +672,6 @@ class CaldavStorage(DavStorage):
     fileext = '.ics'
     item_mimetype = 'text/calendar'
     discovery_class = CalDiscover
-    _dav_namespace = 'urn:ietf:params:xml:ns:caldav'
-    _dav_resourcetype = 'calendar'
 
     start_date = None
     end_date = None
@@ -603,6 +688,11 @@ class CaldavStorage(DavStorage):
 
     get_multi_data_query = '{urn:ietf:params:xml:ns:caldav}calendar-data'
 
+    _property_table = dict(DavStorage._property_table)
+    _property_table.update({
+        'color': ('calendar-color', 'http://apple.com/ns/ical/'),
+    })
+
     def __init__(self, start_date=None, end_date=None,
                  item_types=(), **kwargs):
         super(CaldavStorage, self).__init__(**kwargs)
@@ -683,7 +773,7 @@ class CaldavStorage(DavStorage):
             </C:calendar-query>'''
 
         headers = self.session.get_default_headers()
-        # https://github.com/untitaker/vdirsyncer/issues/166
+        # https://github.com/pimutils/vdirsyncer/issues/166
         # The default in CalDAV's calendar-queries is 0, but the examples use
         # an explicit value of 1 for querying items. it is extremely unclear in
         # the spec which values from WebDAV are actually allowed.
@@ -720,6 +810,4 @@ class CarddavStorage(DavStorage):
                 {hrefs}
             </C:addressbook-multiget>'''
 
-    _dav_namespace = 'urn:ietf:params:xml:ns:carddav'
-    _dav_resourcetype = 'addressbook'
     get_multi_data_query = '{urn:ietf:params:xml:ns:carddav}address-data'
