@@ -5,7 +5,6 @@ from .. import exceptions
 from ..utils import expand_path
 from ..utils.compat import iteritems, text_type, urlparse
 from ..utils.http import request
-from ..utils.password import get_password
 from ..utils.vobject import split_collection
 
 USERAGENT = 'vdirsyncer'
@@ -52,11 +51,11 @@ def prepare_verify(verify, verify_fingerprint):
             raise exceptions.UserError('Invalid value for verify_fingerprint '
                                        '({}), must be a string or null.'
                                        .format(verify_fingerprint))
-        verify = False
     elif not verify:
-        raise exceptions.UserError('verify = false is forbidden. Consider '
-                                   'setting verify_fingerprint instead, which '
-                                   'also disables validation against CAs.')
+        raise exceptions.UserError(
+            'Disabling all SSL validation is forbidden. Consider setting '
+            'verify_fingerprint if you have a broken or self-signed cert.'
+        )
 
     return {
         'verify': verify,
@@ -123,9 +122,6 @@ class HttpStorage(Storage):
                  **kwargs):
         super(HttpStorage, self).__init__(**kwargs)
 
-        if username and not password:
-            password = get_password(username, url)
-
         self._settings = {
             'auth': prepare_auth(auth, username, password),
             'cert': prepare_client_cert(auth_cert),
