@@ -3,17 +3,18 @@
 import collections
 import contextlib
 import functools
+import logging
 import os
 
 from atomicwrites import atomic_write
 
 from .base import Item, Storage
-from .. import exceptions, log
+from .. import exceptions
 from ..utils import checkfile, expand_path
 from ..utils.compat import iteritems, itervalues
 from ..utils.vobject import join_collection, split_collection
 
-logger = log.get(__name__)
+logger = logging.getLogger(__name__)
 
 
 def _writing_op(f):
@@ -37,13 +38,12 @@ class SingleFileStorage(Storage):
 
     .. note::
         This storage is very slow, and that is unlikely to change. You should
-        consider using :py:class:`vdirsyncer.storage.FilesystemStorage` if it
-        fits your usecase.
+        consider using :storage:`filesystem` if it fits your usecase.
 
     :param path: The filepath to the file to be written to.
     :param encoding: Which encoding the file should use. Defaults to UTF-8.
 
-    Example for syncing with :py:class:`vdirsyncer.storage.CaldavStorage`::
+    Example for syncing with :storage:`caldav`::
 
         [pair my_calendar]
         a = my_calendar_local
@@ -134,7 +134,7 @@ class SingleFileStorage(Storage):
     def upload(self, item):
         href = item.ident
         if href in self._items:
-            raise exceptions.AlreadyExistingError(href)
+            raise exceptions.AlreadyExistingError(existing_href=href)
 
         self._items[href] = item, item.hash
         return href, item.hash
