@@ -10,11 +10,13 @@ The algorithm is based on the blogpost "How OfflineIMAP works" by Edward Z.
 Yang. http://blog.ezyang.com/2012/08/how-offlineimap-works/
 '''
 import itertools
+import logging
 
-from . import exceptions, log
+from . import exceptions
 from .utils import uniq
 from .utils.compat import iteritems, text_type
-sync_logger = log.get(__name__)
+
+sync_logger = logging.getLogger(__name__)
 
 
 class SyncError(exceptions.Error):
@@ -81,50 +83,99 @@ class StorageSyncer(object):
     things.'''
     def __init__(self, storage, status):
         '''
-        :param status: {ident: (href, etag)}
+        :param status: {ident: {'href': href, 'etag': etag}}
         '''
         self.storage = storage
         self.status = status
         self.idents = None
 
-    def prepare_idents(self, other_read_only):
-        href_to_status = dict((href, (ident, etag))
-                              for ident, (href, etag)
+    def prepare_idents(self):
+        href_to_status = dict((meta['href'], (ident, meta))
+                              for ident, meta
                               in iteritems(self.status))
 
-        hrefs_to_download = []
+        prefetch = {}
         self.idents = {}
 
-        for href, etag in self.storage.list():
-            if href in href_to_status:
-                ident, old_etag = href_to_status[href]
-                self.idents[ident] = {
-                    'etag': etag,
-                    'href': href,
-                    'ident': ident
-                }
+        def _store_props(ident, props):
+            if self.idents.setdefault(ident, props) is not props:
+                raise IdentConflict(storage=self.storage,
+                                    hrefs=[self.idents[ident]['href'],
+                                           href])
 
-                if etag != old_etag and not other_read_only:
-                    hrefs_to_download.append(href)
+            if ident in self.status:
+                # Necessary if item's href changes.
+                # Otherwise it's a no-op, like `a = a`.
+                self.status[ident]['href'] = props['href']
+
+        for href, etag in self.storage.list():
+            props = {'href': href, 'etag': etag}
+            ident, old_meta = href_to_status.get(href, (None, None))
+            assert etag is not None
+            if old_meta is None or etag != old_meta['etag']:
+                # Either the item is completely new, or updated
+                # In both cases we should prefetch
+                prefetch[href] = props
             else:
-                hrefs_to_download.append(href)
+                _store_props(ident, props)
 
         # Prefetch items
-        for href, item, etag in (self.storage.get_multi(hrefs_to_download) if
-                                 hrefs_to_download else ()):
-            props = self.idents.setdefault(item.ident, {})
-            props['item'] = item
-            props['ident'] = item.ident
+        for href, item, etag in (self.storage.get_multi(prefetch)
+                                 if prefetch else ()):
+            props = prefetch[href]
 
-            if props.setdefault('href', href) != href:
-                raise IdentConflict(storage=self.storage,
-                                    hrefs=[props['href'], href])
-            if props.setdefault('etag', etag) != etag:
-                raise SyncError('Etag changed during sync.')
+            assert props['href'] == href
+            if props['etag'] != etag:
+                sync_logger.warning(
+                    'Etag of {!r} changed during sync from {!r} to {!r}'
+                    .format(href, props['etag'], etag)
+                )
+            props['etag'] = etag
+            props['item'] = item
+            props['ident'] = ident = item.ident
+            _store_props(ident, props)
 
     def is_changed(self, ident):
-        _, status_etag = self.status.get(ident, (None, None))
-        return self.idents[ident]['etag'] != status_etag
+        status = self.status.get(ident, None)
+        meta = self.idents[ident]
+
+        if status is None:  # new item
+            return True
+
+        if meta['etag'] != status['etag']:  # etag changed
+            old_hash = status.get('hash')
+            if old_hash is None or meta['item'].hash != old_hash:
+                # item actually changed
+                return True
+            else:
+                # only etag changed
+                status['etag'] = meta['etag']
+                return False
+
+
+def _status_migrate(status):
+    for ident in list(status):
+        value = status[ident]
+        if len(value) == 4:
+            href_a, etag_a, href_b, etag_b = value
+
+            status[ident] = ({
+                'href': href_a,
+                'etag': etag_a,
+            }, {
+                'href': href_b,
+                'etag': etag_b,
+            })
+
+
+def _compress_meta(meta):
+    '''Make in-memory metadata suitable for disk storage by removing fetched
+    item content'''
+    return {
+        'href': meta['href'],
+        'etag': meta['etag'],
+        'hash': meta['item'].hash
+    }
 
 
 def sync(storage_a, storage_b, status, conflict_resolution=None,
@@ -150,17 +201,19 @@ def sync(storage_a, storage_b, status, conflict_resolution=None,
     if storage_a.read_only and storage_b.read_only:
         raise BothReadOnly()
 
+    _status_migrate(status)
+
     a_info = storage_a.syncer_class(storage_a, dict(
-        (ident, (href_a, etag_a))
-        for ident, (href_a, etag_a, href_b, etag_b) in iteritems(status)
+        (ident, meta_a)
+        for ident, (meta_a, meta_b) in iteritems(status)
     ))
     b_info = storage_b.syncer_class(storage_b, dict(
-        (ident, (href_b, etag_b))
-        for ident, (href_a, etag_a, href_b, etag_b) in iteritems(status)
+        (ident, meta_b)
+        for ident, (meta_a, meta_b) in iteritems(status)
     ))
 
-    a_info.prepare_idents(storage_b.read_only)
-    b_info.prepare_idents(storage_a.read_only)
+    a_info.prepare_idents()
+    b_info.prepare_idents()
 
     if bool(a_info.idents) != bool(b_info.idents) \
        and status and not force_delete:
@@ -176,15 +229,13 @@ def sync(storage_a, storage_b, status, conflict_resolution=None,
 
     status.clear()
     for ident in uniq(itertools.chain(a_info.status, b_info.status)):
-        href_a, etag_a = a_info.status[ident]
-        href_b, etag_b = b_info.status[ident]
-        status[ident] = href_a, etag_a, href_b, etag_b
+        status[ident] = a_info.status[ident], b_info.status[ident]
 
 
 def _action_upload(ident, source, dest):
 
     def inner(a, b, conflict_resolution):
-        sync_logger.info('Copying (uploading) item {} to {}'
+        sync_logger.info(u'Copying (uploading) item {} to {}'
                          .format(ident, dest.storage))
         source_meta = source.idents[ident]
 
@@ -196,8 +247,12 @@ def _action_upload(ident, source, dest):
             item = source_meta['item']
             dest_href, dest_etag = dest.storage.upload(item)
 
-        source.status[ident] = source_meta['href'], source_meta['etag']
-        dest.status[ident] = dest_href, dest_etag
+        source.status[ident] = _compress_meta(source_meta)
+        dest.status[ident] = {
+            'href': dest_href,
+            'hash': source_meta['item'].hash,
+            'etag': dest_etag
+        }
 
     return inner
 
@@ -205,13 +260,13 @@ def _action_upload(ident, source, dest):
 def _action_update(ident, source, dest):
 
     def inner(a, b, conflict_resolution):
-        sync_logger.info('Copying (updating) item {} to {}'
+        sync_logger.info(u'Copying (updating) item {} to {}'
                          .format(ident, dest.storage))
         source_meta = source.idents[ident]
 
         if dest.storage.read_only:
-            sync_logger.info('{dest} is read-only. Skipping update...'
-                             .format(dest=dest.storage))
+            sync_logger.warning(u'{dest} is read-only. Skipping update...'
+                                .format(dest=dest.storage))
             dest_href = dest_etag = None
         else:
             dest_meta = dest.idents[ident]
@@ -220,8 +275,12 @@ def _action_update(ident, source, dest):
                                             dest_meta['etag'])
             assert isinstance(dest_etag, (bytes, text_type))
 
-        source.status[ident] = source_meta['href'], source_meta['etag']
-        dest.status[ident] = dest_href, dest_etag
+        source.status[ident] = _compress_meta(source_meta)
+        dest.status[ident] = {
+            'href': dest_href,
+            'hash': source_meta['item'].hash,
+            'etag': dest_etag
+        }
 
     return inner
 
@@ -231,7 +290,7 @@ def _action_delete(ident, info):
     idents = info.idents
 
     def inner(a, b, conflict_resolution):
-        sync_logger.info('Deleting item {} from {}'.format(ident, storage))
+        sync_logger.info(u'Deleting item {} from {}'.format(ident, storage))
         if storage.read_only:
             sync_logger.warning('{} is read-only, skipping deletion...'
                                 .format(storage))
@@ -249,7 +308,7 @@ def _action_delete(ident, info):
 
 def _action_delete_status(ident):
     def inner(a, b, conflict_resolution):
-        sync_logger.info('Deleting status info for nonexisting item {}'
+        sync_logger.info(u'Deleting status info for nonexisting item {}'
                          .format(ident))
         del a.status[ident]
         del b.status[ident]
@@ -259,23 +318,23 @@ def _action_delete_status(ident):
 
 def _action_conflict_resolve(ident):
     def inner(a, b, conflict_resolution):
-        sync_logger.info('Doing conflict resolution for item {}...'
+        sync_logger.info(u'Doing conflict resolution for item {}...'
                          .format(ident))
         meta_a = a.idents[ident]
         meta_b = b.idents[ident]
 
-        if meta_a['item'].raw == meta_b['item'].raw:
-            sync_logger.info('...same content on both sides.')
-            a.status[ident] = meta_a['href'], meta_a['etag']
-            b.status[ident] = meta_b['href'], meta_b['etag']
+        if meta_a['item'].hash == meta_b['item'].hash:
+            sync_logger.info(u'...same content on both sides.')
+            a.status[ident] = _compress_meta(meta_a)
+            b.status[ident] = _compress_meta(meta_b)
         elif conflict_resolution is None:
             raise SyncConflict(ident=ident, href_a=meta_a['href'],
                                href_b=meta_b['href'])
         elif conflict_resolution == 'a wins':
-            sync_logger.info('...{} wins.'.format(a.storage))
+            sync_logger.info(u'...{} wins.'.format(a.storage))
             _action_update(ident, a, b)(a, b, conflict_resolution)
         elif conflict_resolution == 'b wins':
-            sync_logger.info('...{} wins.'.format(b.storage))
+            sync_logger.info(u'...{} wins.'.format(b.storage))
             _action_update(ident, b, a)(a, b, conflict_resolution)
         else:
             raise exceptions.UserError('Invalid conflict resolution mode: {}'
