@@ -1,12 +1,19 @@
 # -*- coding: utf-8 -*-
 
+import functools
 import os
 import sys
+import uuid
 
-from .compat import iteritems
+from .compat import getargspec_ish, iteritems, to_unicode
 from .. import exceptions
 
 
+SAFE_UID_CHARS = ('abcdefghijklmnopqrstuvwxyz'
+                  'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
+                  '0123456789_.-+')
+
+
 _missing = object()
 
 
@@ -46,46 +53,46 @@ def uniq(s):
             yield x
 
 
-def get_etag_from_file(fpath):
-    '''Get mtime-based etag from a filepath.'''
-    stat = os.stat(fpath)
+def get_etag_from_file(f):
+    '''Get mtime-based etag from a filepath or file-like object.
+
+    This function will flush/sync the file as much as necessary to obtain a
+    correct mtime.
+    '''
+    if hasattr(f, 'read'):
+        f.flush()  # Only this is necessary on Linux
+        if sys.platform == 'win32':
+            os.fsync(f.fileno())  # Apparently necessary on Windows
+        stat = os.fstat(f.fileno())
+    else:
+        stat = os.stat(f)
+
     mtime = getattr(stat, 'st_mtime_ns', None)
     if mtime is None:
         mtime = stat.st_mtime
     return '{:.9f}'.format(mtime)
 
 
-def get_etag_from_fileobject(f):
-    '''
-    Get mtime-based etag from a local file's fileobject.
-
-    This function will flush/sync the file as much as necessary to obtain a
-    correct mtime.
-
-    In filesystem-based storages, this is used instead of
-    :py:func:`get_etag_from_file` to determine the correct etag *before*
-    writing the temporary file to the target location.
-
-    This works because, as far as I've tested, moving and linking a file
-    doesn't change its mtime.
-    '''
-    f.flush()  # Only this is necessary on Linux
-    if sys.platform == 'win32':
-        os.fsync(f.fileno())  # Apparently necessary on Windows
-    return get_etag_from_file(f.name)
-
-
-def get_class_init_specs(cls, stop_at=object):
+def get_storage_init_specs(cls, stop_at=object):
     if cls is stop_at:
         return ()
-    import inspect
-    spec = inspect.getargspec(cls.__init__)
-    supercls = next(getattr(x.__init__, '__objclass__', x)
-                    for x in cls.__mro__[1:])
-    return (spec,) + get_class_init_specs(supercls, stop_at=stop_at)
+
+    spec = getargspec_ish(cls.__init__)
+    traverse_superclass = getattr(cls.__init__, '_traverse_superclass', True)
+    if traverse_superclass:
+        if traverse_superclass is True:  # noqa
+            supercls = next(getattr(x.__init__, '__objclass__', x)
+                            for x in cls.__mro__[1:])
+        else:
+            supercls = traverse_superclass
+        superspecs = get_storage_init_specs(supercls, stop_at=stop_at)
+    else:
+        superspecs = ()
+
+    return (spec,) + superspecs
 
 
-def get_class_init_args(cls, stop_at=object):
+def get_storage_init_args(cls, stop_at=object):
     '''
     Get args which are taken during class initialization. Assumes that all
     classes' __init__ calls super().__init__ with the rest of the arguments.
@@ -96,9 +103,10 @@ def get_class_init_args(cls, stop_at=object):
         requires.
     '''
     all, required = set(), set()
-    for spec in get_class_init_specs(cls, stop_at=stop_at):
+    for spec in get_storage_init_specs(cls, stop_at=stop_at):
         all.update(spec.args[1:])
-        required.update(spec.args[1:-len(spec.defaults or ())])
+        last = -len(spec.defaults) if spec.defaults else len(spec.args)
+        required.update(spec.args[1:last])
 
     return all, required
 
@@ -156,3 +164,59 @@ class cached_property(object):
             return self
         obj.__dict__[self.__name__] = result = self.fget(obj)
         return result
+
+
+def href_safe(ident, safe=SAFE_UID_CHARS):
+    return not bool(set(ident) - set(safe))
+
+
+def generate_href(ident=None, safe=SAFE_UID_CHARS):
+    '''
+    Generate a safe identifier, suitable for URLs, storage hrefs or UIDs.
+
+    If the given ident string is safe, it will be returned, otherwise a random
+    UUID.
+    '''
+    if not ident or not href_safe(ident, safe):
+        return to_unicode(uuid.uuid4().hex)
+    else:
+        return ident
+
+
+def synchronized(lock=None):
+    if lock is None:
+        from threading import Lock
+        lock = Lock()
+
+    def inner(f):
+        @functools.wraps(f)
+        def wrapper(*args, **kwargs):
+            with lock:
+                return f(*args, **kwargs)
+        return wrapper
+    return inner
+
+
+def open_graphical_browser(url, new=0, autoraise=True):
+    '''Open a graphical web browser.
+
+    This is basically like `webbrowser.open`, but without trying to launch CLI
+    browsers at all. We're excluding those since it's undesirable to launch
+    those when you're using vdirsyncer on a server. Rather copypaste the URL
+    into the local browser, or use the URL-yanking features of your terminal
+    emulator.
+    '''
+    import webbrowser
+    cli_names = set(['www-browser', 'links', 'links2', 'elinks', 'lynx',
+                     'w3m'])
+
+    for name in webbrowser._tryorder:
+        if name in cli_names:
+            continue
+
+        browser = webbrowser.get(name)
+        if browser.open(url, new, autoraise):
+            return
+
+    raise RuntimeError('No graphical browser found. Please open the URL '
+                       'manually.')
