@@ -4,7 +4,7 @@ import hashlib
 from itertools import chain, tee
 
 from . import cached_property, uniq
-from .compat import text_type
+from .compat import itervalues, text_type
 
 
 def _process_properties(*s):
@@ -18,14 +18,25 @@ def _process_properties(*s):
 IGNORE_PROPS = _process_properties(
     # PRODID is changed by radicale for some reason after upload
     'PRODID',
-    # VERSION can get lost in singlefile storage
-    'VERSION',
     # X-RADICALE-NAME is used by radicale, because hrefs don't really exist in
     # their filesystem backend
     'X-RADICALE-NAME',
+    # Apparently this is set by Horde?
+    # https://github.com/pimutils/vdirsyncer/issues/318
+    'X-WR-CALNAME',
     # REV is from the VCARD specification and is supposed to change when the
     # item does -- however, we can determine that ourselves
-    'REV'
+    # Same with LAST-MODIFIED
+    'REV',
+    'LAST-MODIFIED',
+    # Added those because e.g. http://www.feiertage-oesterreich.at/ is
+    # generating those randomly on every request.
+    # Some iCalendar HTTP calendars (Google's read-only calendar links)
+    # generate the DTSTAMP at request time, so this property always changes
+    # when the rest of the item didn't.
+    # Some do the same with the UID.
+    'DTSTAMP',
+    'UID',
 )
 del _process_properties
 
@@ -103,17 +114,24 @@ def hash_item(text):
 def split_collection(text):
     assert isinstance(text, text_type)
     inline = []
-    items = []
+    items = {}  # uid => item
+    ungrouped_items = []
 
     def inner(item, main):
         if item.name == u'VTIMEZONE':
             inline.append(item)
         elif item.name == u'VCARD':
-            items.append(item)
+            ungrouped_items.append(item)
         elif item.name in (u'VTODO', u'VEVENT', u'VJOURNAL'):
-            items.append(_Component(main.name,
-                                    main.props[:],
-                                    [item]))
+            uid = item.get(u'UID', u'')
+            wrapper = _Component(main.name, main.props[:], [])
+
+            if uid.strip():
+                wrapper = items.setdefault(uid, wrapper)
+            else:
+                ungrouped_items.append(wrapper)
+
+            wrapper.subcomponents.append(item)
         elif item.name in (u'VCALENDAR', u'VADDRESSBOOK'):
             for subitem in item.subcomponents:
                 inner(subitem, item)
@@ -124,7 +142,7 @@ def split_collection(text):
     for main in _Component.parse(text, multiple=True):
         inner(main, main)
 
-    for item in items:
+    for item in chain(itervalues(items), ungrouped_items):
         item.subcomponents.extend(inline)
         yield u'\r\n'.join(item.dump_lines())
 
@@ -228,19 +246,23 @@ class _Component(object):
 
         stack = []
         rv = []
-        for line in lines:
-            if line.startswith(u'BEGIN:'):
-                c_name = line[len(u'BEGIN:'):].strip().upper()
-                stack.append(cls(c_name, [], []))
-            elif line.startswith(u'END:'):
-                component = stack.pop()
-                if stack:
-                    stack[-1].subcomponents.append(component)
+        try:
+            for i, line in enumerate(lines):
+                if line.startswith(u'BEGIN:'):
+                    c_name = line[len(u'BEGIN:'):].strip().upper()
+                    stack.append(cls(c_name, [], []))
+                elif line.startswith(u'END:'):
+                    component = stack.pop()
+                    if stack:
+                        stack[-1].subcomponents.append(component)
+                    else:
+                        rv.append(component)
                 else:
-                    rv.append(component)
-            else:
-                if line.strip():
-                    stack[-1].props.append(line)
+                    if line.strip():
+                        stack[-1].props.append(line)
+        except IndexError:
+            raise ValueError('Parsing error at line {}. Check the debug log '
+                             'for more information.'.format(i + 1))
 
         if multiple:
             return rv
@@ -278,6 +300,8 @@ class _Component(object):
         self.props = new_lines
 
     def __setitem__(self, key, val):
+        assert isinstance(val, text_type)
+        assert u'\n' not in val
         del self[key]
         line = u'{}:{}'.format(key, val)
         self.props.append(line)
@@ -299,3 +323,9 @@ class _Component(object):
                 break
 
         return rv
+
+    def get(self, key, default=None):
+        try:
+            return self[key]
+        except KeyError:
+            return default
